@@ -16,14 +16,17 @@
 #   identifier to be stored in the converted revision. This will cause
 #   the converted revision to have a different identity than the
 #   source.
-from __future__ import absolute_import
+
+from __future__ import annotations
 
 import os
 import re
 import time
 
 from mercurial.i18n import _
-from mercurial.pycompat import open
+from mercurial.interfaces.types import (
+    RepoT,
+)
 from mercurial.node import (
     bin,
     hex,
@@ -34,14 +37,14 @@ from mercurial import (
     context,
     error,
     exchange,
-    hg,
     lock as lockmod,
     logcmdutil,
     merge as mergemod,
+    mergestate,
     phases,
-    pycompat,
     util,
 )
+from mercurial.repo import factory as repo_factory
 from mercurial.utils import dateutil
 
 stringio = util.stringio
@@ -63,7 +66,7 @@ class mercurial_sink(common.converter_sink):
         self.lastbranch = None
         if os.path.isdir(path) and len(os.listdir(path)) > 0:
             try:
-                self.repo = hg.repository(self.ui, path)
+                self.repo = repo_factory.repository(self.ui, path)
                 if not self.repo.local():
                     raise NoRepo(
                         _(b'%s is not a local Mercurial repository') % path
@@ -74,7 +77,7 @@ class mercurial_sink(common.converter_sink):
         else:
             try:
                 ui.status(_(b'initializing destination %s repository\n') % path)
-                self.repo = hg.repository(self.ui, path, create=True)
+                self.repo = repo_factory.repository(self.ui, path, create=True)
                 if not self.repo.local():
                     raise NoRepo(
                         _(b'%s is not a local Mercurial repository') % path
@@ -122,9 +125,13 @@ class mercurial_sink(common.converter_sink):
         if setbranch:
             self.after()
             try:
-                self.repo = hg.repository(self.ui, branchpath)
+                self.repo = repo_factory.repository(self.ui, branchpath)
             except Exception:
-                self.repo = hg.repository(self.ui, branchpath, create=True)
+                self.repo = repo_factory.repository(
+                    self.ui,
+                    branchpath,
+                    create=True,
+                )
             self.before()
 
         # pbranches may bring revisions from other branches (merge parents)
@@ -138,9 +145,9 @@ class mercurial_sink(common.converter_sink):
 
         if missings:
             self.after()
-            for pbranch, heads in sorted(pycompat.iteritems(missings)):
+            for pbranch, heads in sorted(missings.items()):
                 pbranchpath = os.path.join(self.path, pbranch)
-                prepo = hg.peer(self.ui, {}, pbranchpath)
+                prepo = repo_factory.peer(self.ui, {}, pbranchpath)
                 self.ui.note(
                     _(b'pulling from %s into %s\n') % (pbranch, branch)
                 )
@@ -241,7 +248,7 @@ class mercurial_sink(common.converter_sink):
 
             # If the file requires actual merging, abort. We don't have enough
             # context to resolve merges correctly.
-            if action in [b'm', b'dm', b'cd', b'dc']:
+            if action in mergestate.CONVERT_MERGE_ACTIONS:
                 raise error.Abort(
                     _(
                         b"unable to convert merge commit "
@@ -250,7 +257,7 @@ class mercurial_sink(common.converter_sink):
                     )
                     % (file, p1ctx, p2ctx)
                 )
-            elif action == b'k':
+            elif action == mergestate.ACTION_KEEP:
                 # 'keep' means nothing changed from p1
                 continue
             else:
@@ -299,8 +306,9 @@ class mercurial_sink(common.converter_sink):
         parents = pl
         nparents = len(parents)
         if self.filemapmode and nparents == 1:
-            m1node = self.repo.changelog.read(bin(parents[0]))[0]
             parent = parents[0]
+            p1_node = bin(parent)
+            m1node = self.repo.changelog.changelogrevision(p1_node).manifest
 
         if len(parents) < 2:
             parents.append(self.repo.nullid)
@@ -423,8 +431,9 @@ class mercurial_sink(common.converter_sink):
         tagparent = tagparent or self.repo.nullid
 
         oldlines = set()
-        for branch, heads in pycompat.iteritems(self.repo.branchmap()):
-            for h in heads:
+        bm = self.repo.branchmap()
+        for branch in bm:
+            for h in bm.branchheads(branch, closed=True):
                 if b'.hgtags' in self.repo[h]:
                     oldlines.update(
                         set(self.repo[h][b'.hgtags'].data().splitlines(True))
@@ -520,7 +529,7 @@ class mercurial_source(common.converter_source):
         self.ignored = set()
         self.saverev = ui.configbool(b'convert', b'hg.saverev')
         try:
-            self.repo = hg.repository(self.ui, path)
+            self.repo: RepoT = repo_factory.repository(self.ui, path)
             # try to provoke an exception if this isn't really a hg
             # repo, but some other bogus compatible-looking url
             if not self.repo.local():
@@ -595,7 +604,7 @@ class mercurial_source(common.converter_source):
         maappend = ma.append
         rappend = r.append
         d = ctx1.manifest().diff(ctx2.manifest())
-        for f, ((node1, flag1), (node2, flag2)) in pycompat.iteritems(d):
+        for f, ((node1, flag1), (node2, flag2)) in d.items():
             if node2 is None:
                 rappend(f)
             else:
@@ -609,7 +618,10 @@ class mercurial_source(common.converter_source):
             files = copyfiles = ctx.manifest()
         if parents:
             if self._changescache[0] == rev:
-                ma, r = self._changescache[1]
+                # TODO: add type hints to avoid this warning, instead of
+                #  suppressing it:
+                #     No attribute '__iter__' on None [attribute-error]
+                ma, r = self._changescache[1]  # pytype: disable=attribute-error
             else:
                 ma, r = self._changedfiles(parents[0], ctx)
             if not full:
@@ -621,7 +633,7 @@ class mercurial_source(common.converter_source):
         cleanp2 = set()
         if len(parents) == 2:
             d = parents[1].manifest().diff(ctx.manifest(), clean=True)
-            for f, value in pycompat.iteritems(d):
+            for f, value in d.items():
                 if value is None:
                     cleanp2.add(f)
         changes = [(f, rev) for f in files if f not in self.ignored]
@@ -693,7 +705,7 @@ class mercurial_source(common.converter_source):
         else:
             i = i or 0
             ma, r = self._changedfiles(parents[i], ctx)
-        ma, r = [[f for f in l if f not in self.ignored] for l in (ma, r)]
+        ma, r = ([f for f in l if f not in self.ignored] for l in (ma, r))
 
         if i == 0:
             self._changescache = (rev, (ma, r))
@@ -702,7 +714,7 @@ class mercurial_source(common.converter_source):
 
     def converted(self, rev, destrev):
         if self.convertfp is None:
-            self.convertfp = open(self.repo.vfs.join(b'shamap'), b'ab')
+            self.convertfp = open(self.repo.vfs.join(b'shamap'), 'ab')
         self.convertfp.write(util.tonativeeol(b'%s %s\n' % (destrev, rev)))
         self.convertfp.flush()
 
