@@ -190,7 +190,7 @@ unexpectedly::
 
 """
 
-from __future__ import absolute_import
+from __future__ import annotations
 
 # chistedit dependencies that are not available everywhere
 try:
@@ -200,15 +200,13 @@ except ImportError:
     fcntl = None
     termios = None
 
+import binascii
 import functools
 import os
+import pickle
 import struct
 
 from mercurial.i18n import _
-from mercurial.pycompat import (
-    getattr,
-    open,
-)
 from mercurial.node import (
     bin,
     hex,
@@ -225,7 +223,6 @@ from mercurial import (
     error,
     exchange,
     extensions,
-    hg,
     logcmdutil,
     merge as mergemod,
     mergestate as mergestatemod,
@@ -239,13 +236,21 @@ from mercurial import (
     state as statemod,
     util,
 )
+from mercurial.cmd_impls import (
+    update as up_impl,
+)
+from mercurial.merge_utils import (
+    update as update_util,
+)
+from mercurial.repo import (
+    factory as repo_factory,
+)
 from mercurial.utils import (
     dateutil,
     stringutil,
     urlutil,
 )
 
-pickle = util.pickle
 cmdtable = {}
 command = registrar.command(cmdtable)
 
@@ -352,7 +357,7 @@ Commands:
     return b''.join([b'# %s\n' % l if l else b'#\n' for l in lines])
 
 
-class histeditstate(object):
+class histeditstate:
     def __init__(self, repo):
         self.repo = repo
         self.actions = None
@@ -455,7 +460,7 @@ class histeditstate(object):
         rules = []
         rulelen = int(lines[index])
         index += 1
-        for i in pycompat.xrange(rulelen):
+        for i in range(rulelen):
             ruleaction = lines[index]
             index += 1
             rule = lines[index]
@@ -466,7 +471,7 @@ class histeditstate(object):
         replacements = []
         replacementlen = int(lines[index])
         index += 1
-        for i in pycompat.xrange(replacementlen):
+        for i in range(replacementlen):
             replacement = lines[index]
             original = bin(replacement[:40])
             succ = [
@@ -491,7 +496,7 @@ class histeditstate(object):
         return self.repo.vfs.exists(b'histedit-state')
 
 
-class histeditaction(object):
+class histeditaction:
     def __init__(self, state, node):
         self.state = state
         self.repo = state.repo
@@ -505,7 +510,7 @@ class histeditaction(object):
         # Check for validation of rule ids and get the rulehash
         try:
             rev = bin(ruleid)
-        except TypeError:
+        except binascii.Error:
             try:
                 _ctx = scmutil.revsingle(state.repo, ruleid)
                 rulehash = _ctx.hex()
@@ -553,9 +558,7 @@ class histeditaction(object):
             summary = cmdutil.rendertemplate(
                 ctx, ui.config(b'histedit', b'summary-template')
             )
-        # Handle the fact that `''.splitlines() => []`
-        summary = summary.splitlines()[0] if summary else b''
-        line = b'%s %s %s' % (self.verb, ctx, summary)
+        line = b'%s %s %s' % (self.verb, ctx, stringutil.firstline(summary))
         # trim to 75 columns by default so it's not stupidly wide in my editor
         # (the 5 more are left for verb)
         maxlen = self.repo.ui.configint(b'histedit', b'linelen')
@@ -581,9 +584,9 @@ class histeditaction(object):
         repo = self.repo
         rulectx = repo[self.node]
         with repo.ui.silent():
-            hg.update(repo, self.state.parentctxnode, quietempty=True)
+            up_impl.update(repo, self.state.parentctxnode, quietempty=True)
         stats = applychanges(repo.ui, repo, rulectx, {})
-        repo.dirstate.setbranch(rulectx.branch())
+        repo.dirstate.setbranch(rulectx.branch(), repo.currenttransaction())
         if stats.unresolvedcount:
             raise error.InterventionRequired(
                 _(b'Fix up the change (%s %s)') % (self.verb, short(self.node)),
@@ -660,14 +663,22 @@ def applychanges(ui, repo, ctx, opts):
         # just applies changes on parent for editing
         with ui.silent():
             cmdutil.revert(ui, repo, ctx, all=True)
-            stats = mergemod.updateresult(0, 0, 0, 0)
+            stats = update_util.UpdateResult(0, 0, 0, 0)
     else:
         try:
             # ui.forcemerge is an internal variable, do not document
             repo.ui.setconfig(
                 b'ui', b'forcemerge', opts.get(b'tool', b''), b'histedit'
             )
-            stats = mergemod.graft(repo, ctx, labels=[b'local', b'histedit'])
+            stats = mergemod.graft(
+                repo,
+                ctx,
+                labels=[
+                    b'already edited',
+                    b'current change',
+                    b'parent of current change',
+                ],
+            )
         finally:
             repo.ui.setconfig(b'ui', b'forcemerge', b'', b'histedit')
     return stats
@@ -794,7 +805,7 @@ class pick(histeditaction):
             self.repo.ui.debug(b'node %s unchanged\n' % short(self.node))
             return rulectx, []
 
-        return super(pick, self).run()
+        return super().run()
 
 
 @action(
@@ -806,7 +817,7 @@ class edit(histeditaction):
     def run(self):
         repo = self.repo
         rulectx = repo[self.node]
-        hg.update(repo, self.state.parentctxnode, quietempty=True)
+        up_impl.update(repo, self.state.parentctxnode, quietempty=True)
         applychanges(repo.ui, repo, rulectx, {})
         hint = _(b'to edit %s, `hg histedit --continue` after making changes')
         raise error.InterventionRequired(
@@ -823,7 +834,7 @@ class edit(histeditaction):
 class fold(histeditaction):
     def verify(self, prev, expected, seen):
         """Verifies semantic correctness of the fold rule"""
-        super(fold, self).verify(prev, expected, seen)
+        super().verify(prev, expected, seen)
         repo = self.repo
         if not prev:
             c = repo[self.node].p1()
@@ -990,7 +1001,7 @@ class base(histeditaction):
 @action(
     [b'_multifold'],
     _(
-        """fold subclass used for when multiple folds happen in a row
+        b"""fold subclass used for when multiple folds happen in a row
 
     We only want to fire the editor for the folded message once when
     (say) four changes are folded down into a single change. This is
@@ -1045,12 +1056,16 @@ def findoutgoing(ui, repo, remote=None, force=False, opts=None):
     if opts is None:
         opts = {}
     path = urlutil.get_unique_push_path(b'histedit', repo, ui, remote)
-    dest = path.pushloc or path.loc
 
-    ui.status(_(b'comparing with %s\n') % urlutil.hidepassword(dest))
+    ui.status(_(b'comparing with %s\n') % urlutil.hidepassword(path.loc))
 
-    revs, checkout = hg.addbranchrevs(repo, repo, (path.branch, []), None)
-    other = hg.peer(repo, opts, dest)
+    revs, checkout = urlutil.add_branch_revs(
+        repo,
+        repo,
+        (path.branch, []),
+        None,
+    )
+    other = repo_factory.peer(repo, opts, path)
 
     if revs:
         revs = [repo.lookup(rev) for rev in revs]
@@ -1135,7 +1150,7 @@ def screen_size():
     return struct.unpack(b'hh', fcntl.ioctl(1, termios.TIOCGWINSZ, b'    '))
 
 
-class histeditrule(object):
+class histeditrule:
     def __init__(self, ui, ctx, pos, action=b'pick'):
         self.ui = ui
         self.ctx = ctx
@@ -1185,7 +1200,7 @@ class histeditrule(object):
         # This is split off from the prefix property so that we can
         # separately make the description for 'roll' red (since it
         # will get discarded).
-        return self.ctx.description().splitlines()[0].strip()
+        return stringutil.firstline(self.ctx.description())
 
     def checkconflicts(self, other):
         if other.pos > self.pos and other.origpos <= self.origpos:
@@ -1235,7 +1250,7 @@ def _trunc_tail(line, n):
     return line[: n - 2] + b' >'
 
 
-class _chistedit_state(object):
+class _chistedit_state:
     def __init__(
         self,
         repo,
@@ -1284,7 +1299,7 @@ class _chistedit_state(object):
         line = b"bookmark:  %s" % b' '.join(bms)
         win.addstr(3, 1, line[:length])
 
-        line = b"summary:   %s" % (ctx.description().splitlines()[0])
+        line = b"summary:   %s" % stringutil.firstline(ctx.description())
         win.addstr(4, 1, line[:length])
 
         line = b"files:     "
@@ -1324,6 +1339,10 @@ pgup: prev page, space/pgdn: next page, c: commit, q: abort
 d: drop, e: edit, f: fold, m: mess, p: pick, r: roll
 pgup/K: move patch up, pgdn/J: move patch down, c: commit, q: abort
 """
+            if self.later_on_top:
+                help += b"Newer commits are shown above older commits.\n"
+            else:
+                help += b"Older commits are shown above newer commits.\n"
         return help.splitlines()
 
     def render_help(self, win):
@@ -1418,11 +1437,11 @@ pgup/K: move patch up, pgdn/J: move patch down, c: commit, q: abort
         for y in range(0, length):
             line = output[y]
             if diffcolors:
-                if line and line[0] == b'+':
+                if line.startswith(b'+'):
                     win.addstr(
                         y, 0, line, curses.color_pair(COLOR_DIFF_ADD_LINE)
                     )
-                elif line and line[0] == b'-':
+                elif line.startswith(b'-'):
                     win.addstr(
                         y, 0, line, curses.color_pair(COLOR_DIFF_DEL_LINE)
                     )
@@ -1478,7 +1497,7 @@ pgup/K: move patch up, pgdn/J: move patch down, c: commit, q: abort
             self.selected = oldpos if self.selected is None else None
             self.make_selection(self.selected)
         elif action == b'goto' and int(ch) < len(self.rules) <= 10:
-            newrule = next((r for r in self.rules if r.origpos == int(ch)))
+            newrule = next(r for r in self.rules if r.origpos == int(ch))
             self.move_cursor(oldpos, newrule.pos)
             if self.selected is not None:
                 self.swap(oldpos, newrule.pos)
@@ -1518,7 +1537,8 @@ pgup/K: move patch up, pgdn/J: move patch down, c: commit, q: abort
 
     def move_cursor(self, oldpos, newpos):
         """Change the rule/changeset that the cursor is pointing to, regardless of
-        current mode (you can switch between patches from the view patch window)."""
+        current mode (you can switch between patches from the view patch window).
+        """
         self.pos = newpos
 
         mode, _ = self.mode
@@ -1564,7 +1584,7 @@ pgup/K: move patch up, pgdn/J: move patch down, c: commit, q: abort
 
         start = min(old_rule_pos, new_rule_pos)
         end = max(old_rule_pos, new_rule_pos)
-        for r in pycompat.xrange(start, end + 1):
+        for r in range(start, end + 1):
             rules[new_rule_pos].checkconflicts(rules[r])
             rules[old_rule_pos].checkconflicts(rules[r])
 
@@ -1573,14 +1593,16 @@ pgup/K: move patch up, pgdn/J: move patch down, c: commit, q: abort
 
     def change_action(self, pos, action):
         """Change the action state on the given position to the new action"""
-        assert 0 <= pos < len(self.rules)
-        self.rules[pos].action = action
+        rule_pos = self.display_pos_to_rule_pos(pos)
+        assert 0 <= rule_pos < len(self.rules)
+        self.rules[rule_pos].action = action
 
     def cycle_action(self, pos, next=False):
         """Changes the action state the next or the previous action from
         the action list"""
-        assert 0 <= pos < len(self.rules)
-        current = self.rules[pos].action
+        rule_pos = self.display_pos_to_rule_pos(pos)
+        assert 0 <= rule_pos < len(self.rules)
+        current = self.rules[rule_pos].action
 
         assert current in KEY_LIST
 
@@ -1589,11 +1611,14 @@ pgup/K: move patch up, pgdn/J: move patch down, c: commit, q: abort
             index += 1
         else:
             index -= 1
+        # using pos instead of rule_pos because change_action() also calls
+        # display_pos_to_rule_pos()
         self.change_action(pos, KEY_LIST[index % len(KEY_LIST)])
 
     def change_view(self, delta, unit):
         """Change the region of whatever is being viewed (a patch or the list of
-        changesets). 'delta' is an amount (+/- 1) and 'unit' is 'page' or 'line'."""
+        changesets). 'delta' is an amount (+/- 1) and 'unit' is 'page' or 'line'.
+        """
         mode, _ = self.mode
         if mode != MODE_PATCH:
             return
@@ -1702,7 +1727,7 @@ def _chisteditmain(repo, rules, stdscr):
             ch = encoding.strtolocal(stdscr.getkey())
 
 
-def _chistedit(ui, repo, freeargs, opts):
+def _chistedit(ui, repo, state, freeargs, opts):
     """interactively edit changeset history via a curses interface
 
     Provides a ncurses interface to histedit. Press ? in chistedit mode
@@ -1718,7 +1743,7 @@ def _chistedit(ui, repo, freeargs, opts):
         keep = opts.get(b'keep')
         revs = opts.get(b'rev', [])[:]
         cmdutil.checkunfinished(repo)
-        cmdutil.bailifchanged(repo)
+        scmutil.bail_if_changed(repo)
 
         revs.extend(freeargs)
         if not revs:
@@ -1752,8 +1777,6 @@ def _chistedit(ui, repo, freeargs, opts):
             rules.append(histeditrule(ui, repo[r], i))
         with util.with_lc_ctype():
             rc = curses.wrapper(functools.partial(_chisteditmain, repo, rules))
-        curses.echo()
-        curses.endwin()
         if rc is False:
             ui.write(_(b"histedit aborted\n"))
             return 0
@@ -1764,7 +1787,7 @@ def _chistedit(ui, repo, freeargs, opts):
                 for r in rules:
                     fp.write(r)
                 opts[b'commands'] = fp.name
-            return _texthistedit(ui, repo, freeargs, opts)
+            return _texthistedit(ui, repo, state, freeargs, opts)
     except KeyboardInterrupt:
         pass
     return -1
@@ -1903,20 +1926,20 @@ def histedit(ui, repo, *freeargs, **opts):
     """
     opts = pycompat.byteskwargs(opts)
 
-    # kludge: _chistedit only works for starting an edit, not aborting
-    # or continuing, so fall back to regular _texthistedit for those
-    # operations.
-    if ui.interface(b'histedit') == b'curses' and _getgoal(opts) == goalnew:
-        return _chistedit(ui, repo, freeargs, opts)
-    return _texthistedit(ui, repo, freeargs, opts)
-
-
-def _texthistedit(ui, repo, freeargs, opts):
     state = histeditstate(repo)
     with repo.wlock() as wlock, repo.lock() as lock:
         state.wlock = wlock
         state.lock = lock
-        _histedit(ui, repo, state, freeargs, opts)
+        # kludge: _chistedit only works for starting an edit, not aborting
+        # or continuing, so fall back to regular _texthistedit for those
+        # operations.
+        if ui.interface(b'histedit') == b'curses' and _getgoal(opts) == goalnew:
+            return _chistedit(ui, repo, state, freeargs, opts)
+        return _texthistedit(ui, repo, state, freeargs, opts)
+
+
+def _texthistedit(ui, repo, state, freeargs, opts):
+    _histedit(ui, repo, state, freeargs, opts)
 
 
 goalcontinue = b'continue'
@@ -1940,8 +1963,7 @@ def _readfile(ui, path):
         with ui.timeblockedsection(b'histedit'):
             return ui.fin.read()
     else:
-        with open(path, b'rb') as f:
-            return f.read()
+        return util.readfile(path)
 
 
 def _validateargs(ui, repo, freeargs, opts, goal, rules, revs):
@@ -2090,7 +2112,7 @@ def _finishhistedit(ui, repo, state, fm):
 
     mapping, tmpnodes, created, ntm = processreplacement(state)
     if mapping:
-        for prec, succs in pycompat.iteritems(mapping):
+        for prec, succs in mapping.items():
             if not succs:
                 ui.debug(b'histedit: %s is dropped\n' % short(prec))
             else:
@@ -2121,14 +2143,14 @@ def _finishhistedit(ui, repo, state, fm):
         for k, v in mapping.items()
         if has_node(k) and all(has_node(n) for n in v)
     }
-    scmutil.cleanupnodes(repo, mapping, b'histedit')
+    cmdutil.cleanup_nodes(repo, mapping, b'histedit')
     hf = fm.hexfunc
     fl = fm.formatlist
     fd = fm.formatdict
     nodechanges = fd(
         {
             hf(oldn): fl([hf(n) for n in newn], name=b'node')
-            for oldn, newn in pycompat.iteritems(mapping)
+            for oldn, newn in mapping.items()
         },
         key=b"oldnode",
         value=b"newnodes",
@@ -2151,7 +2173,7 @@ def _aborthistedit(ui, repo, state, nobackup=False):
         # Recover our old commits if necessary
         if not state.topmost in repo and state.backupfile:
             backupfile = repo.vfs.join(state.backupfile)
-            f = hg.openpath(ui, backupfile)
+            f = scmutil.open_path(ui, backupfile)
             gen = exchange.readbundle(ui, f, backupfile)
             with repo.transaction(b'histedit.abort') as tr:
                 bundle2.applybundle(
@@ -2170,7 +2192,7 @@ def _aborthistedit(ui, repo, state, nobackup=False):
             state.parentctxnode,
             leafs | tmpnodes,
         ):
-            hg.clean(repo, state.topmost, show_stats=True, quietempty=True)
+            up_impl.clean(repo, state.topmost, show_stats=True, quietempty=True)
         cleanupnode(ui, repo, tmpnodes, nobackup=nobackup)
         cleanupnode(ui, repo, leafs, nobackup=nobackup)
     except Exception:
@@ -2218,7 +2240,7 @@ def _newhistedit(ui, repo, state, revs, freeargs, opts):
     force = opts.get(b'force')
 
     cmdutil.checkunfinished(repo)
-    cmdutil.bailifchanged(repo)
+    scmutil.bail_if_changed(repo)
 
     topmost = repo.dirstate.p1()
     if outg:
@@ -2310,12 +2332,7 @@ def _newhistedit(ui, repo, state, revs, freeargs, opts):
 
 
 def _getsummary(ctx):
-    # a common pattern is to extract the summary but default to the empty
-    # string
-    summary = ctx.description() or b''
-    if summary:
-        summary = summary.splitlines()[0]
-    return summary
+    return stringutil.firstline(ctx.description())
 
 
 def bootstrapcontinue(ui, state, opts):
@@ -2340,6 +2357,21 @@ def bootstrapcontinue(ui, state, opts):
     return state
 
 
+def hgcontinuehistedit(ui, repo):
+    fm = ui.formatter(b'continue', {})
+    fm.startitem()
+
+    state = histeditstate(repo)
+    with repo.wlock() as wlock, repo.lock() as lock:
+        state.wlock = wlock
+        state.lock = lock
+        state.read()
+        state = bootstrapcontinue(ui, state, None)
+        _continuehistedit(ui, repo, state)
+        _finishhistedit(ui, repo, state, fm)
+        fm.end()
+
+
 def between(repo, old, new, keep):
     """select and validate the set of revision to edit
 
@@ -2376,7 +2408,7 @@ def ruleeditor(repo, ui, actions, editcomment=b""):
                     tsum = summary[len(fword) + 1 :].lstrip()
                     # safe but slow: reverse iterate over the actions so we
                     # don't clash on two commits having the same summary
-                    for na, l in reversed(list(pycompat.iteritems(newact))):
+                    for na, l in reversed(list(newact.items())):
                         actx = repo[na.node]
                         asum = _getsummary(actx)
                         if asum == tsum:
@@ -2389,7 +2421,7 @@ def ruleeditor(repo, ui, actions, editcomment=b""):
 
         # copy over and flatten the new list
         actions = []
-        for na, l in pycompat.iteritems(newact):
+        for na, l in newact.items():
             actions.append(na)
             actions += l
 
@@ -2648,7 +2680,7 @@ def stripwrapper(orig, ui, repo, nodelist, *args, **kwargs):
     return orig(ui, repo, nodelist, *args, **kwargs)
 
 
-extensions.wrapfunction(repair, b'strip', stripwrapper)
+extensions.wrapfunction(repair, 'strip', stripwrapper)
 
 
 def summaryhook(ui, repo):
@@ -2675,4 +2707,5 @@ def extsetup(ui):
         allowcommit=True,
         continueflag=True,
         abortfunc=hgaborthistedit,
+        continuefunc=hgcontinuehistedit,
     )
