@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 #
 # A portable replacement for 'seq'
 #
@@ -7,20 +7,15 @@
 #   seq START STOP        [START, STOP] stepping by 1
 #   seq START STEP STOP   [START, STOP] stepping by STEP
 
-from __future__ import absolute_import, print_function
-import os
+import io
 import sys
 
-try:
-    import msvcrt
-    msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
-    msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)
-    msvcrt.setmode(sys.stderr.fileno(), os.O_BINARY)
-except ImportError:
-    pass
-
-if sys.version_info[0] >= 3:
-    xrange = range
+sys.stdout = io.TextIOWrapper(
+    sys.stdout.buffer,
+    sys.stdout.encoding,
+    sys.stdout.errors,
+    newline="\n",
+)
 
 start = 1
 if len(sys.argv) > 2:
@@ -32,5 +27,5 @@ if len(sys.argv) > 3:
 
 stop = int(sys.argv[-1]) + 1
 
-for i in xrange(start, stop, step):
+for i in range(start, stop, step):
     print(i)
